#wiiSocket.py

from wiiMoteErrors import WiiMoteShutdownError
from lightblue import socket as bluetoothSocket, L2CAP, finddevices, BluetoothError
import thread
import time
import socket
import struct

class WiiMote():
    '''represents a WiiMote-Bluetooth-HID as an readable and controllable object '''
    
    def __init__( self, name, address ):
        '''create WiiMote-Object with name and unique mac-address'''
        
        self.msg = ''
        self.message = ''
        self.zeroGX = .0
        self.zeroGY = .0
        self.zeroGZ = .0

        self.name = name
        self.address = address
        self.bluetoothIn = bluetoothSocket( L2CAP )
        self.bluetoothOut = bluetoothSocket( L2CAP )
        print 'new WiiMote'
    
    def connect( self ):
        '''connects the WiiMote
        
        sets up the bluetooth-sockets for in- and output'''
        print 'try to connect',self.address
       
        self.bluetoothIn.connect( ( self.address, 0x13 ) )
        self.bluetoothOut.connect( ( self.address, 0x11 ) )
        self.bluetoothIn.settimeout( 1 )
        self.__receiveAndStoreCalibrationData()
        print 'WiiMote',self.name,'@',self.address,'connected'
        
    def rumble( self, rumbleTime ):
        '''rumbles for the given Time'''
        self.bluetoothOut.send( '\x52\x13\x01' )
        #replace with timer/thread!
        time.sleep( rumbleTime )
        self.bluetoothOut.send( '\x52\x13\x00' )
        self.__receiveMotionData()
        
    def led( self, which ):
        '''leds from 1 to 4 can be switched on, 0 switches all off'''
        leds = [ '\x00', '\x10', '\x20', '\x40', '\x80' ]
        if not( -1 < which < 5 ):
            which = 0
        self.bluetoothOut.send( '\x52\x11' + leds[which] )
        self.__receiveMotionData()
        
    def receive( self ):
        '''receive data from the bluetoothsocket, normalize it and write it into the instance-variable "message"'''
        try:
            self.msg = self.bluetoothIn.recv( 23 )
            if len( self.msg ) > 6:
                self.message = self.msg[2] + self.msg[3];
                
                gX = ( ord ( self.msg[4] )  - self.zeroGX ) / 25.
                gY = ( ord ( self.msg[5] )  - self.zeroGY ) / 25.
                gZ = ( ord ( self.msg[6] )  - self.zeroGZ ) / 25.
                
                self.message += struct.pack( '!f', gX );
                self.message += struct.pack( '!f', gY );
                self.message += struct.pack( '!f', gZ );
            if len( self.msg ) == 0:
                raise WiiMoteShutdownError("WiiMote shut down by user")
        except:
            if len( self.msg ) == 0:
                raise WiiMoteShutdownError("WiiMote shut down by user")
    
    def __receiveAndStoreCalibrationData( self ):
        '''receive calibration-data: zero G for x-, y-, and z-axis'''
        self.bluetoothOut.send( '\x52\x17\x00\x00\x00\x16\x00\x10' )
        while True:
            try:
                self.msg = self.bluetoothIn.recv( 23 )
                if len( self.msg ) > 22:
                   
                    self.zeroGX =  ord( self.msg[7] )
                    self.zeroGY =  ord( self.msg[8] )
                    self.zeroGZ =  ord( self.msg[9] )
                    
                    self.__receiveMotionData()
                    break
            except BluetoothError:
                print 'no wiiMote-calibration-data available'
    
    def __receiveMotionData( self ):
        '''setting the SET_REPORT request to channel 12 for getting motion sensor data
        
        set third byte for continous output to \x04,
        to report only canges to \x00 (recommanded)'''
        self.bluetoothOut.send( '\x52\x12\x00\x31' )
        

#looking for devices
def findWiiThread( wiiMotes ):
    '''thread who find WiiMotes in the bluetooth-enviroment
    
    new wiiMotes are automaticlly attached'''
    while True:
        nearby_devices = finddevices( getnames = True, length = 100 )
        #print "searching bluetooth devices..."
        for address, name, classOfDevice in nearby_devices:
            print name, address, classOfDevice
            if name == 'Nintendo RVL-CNT-01':
                addresses = []
                for wM in wiiMotes:
                    addresses.append( wM.address )
                if not address in addresses:
                    wiiMote = WiiMote( name, address )
                    try:
                        wiiMote.connect()
                        wiiMotes.append( wiiMote )
                        wiiMote.led( len( wiiMotes ) % 4 )
                        thread.start_new_thread( readBluetoothThread, ( wiiMote, wiiMotes ) )
                    except BluetoothError:
                        del wiiMote
                        print 'Error while connecting, next try...'
        time.sleep( 1 )
    
def readBluetoothThread( wiiMote, wiiMotes ):
    '''thread which reads in the data of the given WiiMote
    
    deletes the WiiMote if it is shut down'''
    while True:
        try:
            wiiMote.receive()
        except WiiMoteShutdownError, detail:
            print detail
            wiiMotes.remove( wiiMote )
            del wiiMote
            exit()

def clientThread( clientsocket, WiiMotes ):
    '''thread wich sends the data of all WiiMotes to the given clientSocket'''
    while True:
        msg = ''
        for wiiMote in WiiMotes:
            if len( wiiMote.message ) > 0:
                msg += wiiMote.address + wiiMote.message
        try:
            clientsocket.send( msg )
        except Exception:
            print 'bye bye client-socket', clientsocket
            clientsocket.close()
            exit()
        time.sleep( .02 )
            
def main( host, port, maxConnections ):
    '''WiiSocket, initialise with host(ip), port(number), and the maximum amount of connections
    
    usage (standalone):
    just execute the file, if you whish to use another host, another port
    or another amount of maximal connections, simply adjust the arguments
    in the last line of this document
    
    usage (in your module):
    put this module-file in one of your python-paths or in the same directory
    as your using module.
    >>>import WiiSocket as wiiSocket
    >>>wiiSocket.main(ipOfLocalhost, portNumber, maxAmountOfClients)'''
    
    WiiMotes = []
    thread.start_new_thread(findWiiThread, ( WiiMotes, ) )
    serverSocket = socket.socket( socket.AF_INET, socket.SOCK_STREAM )
    serverSocket.bind( ( host, port ) )
    serverSocket.listen( maxConnections )
    while True:
            #accept connections from outside
            print 'waiting for incoming clients'
            (clientSocket, clientAddress) = serverSocket.accept()
            print 'connected with', clientAddress
            thread.start_new_thread(clientThread, ( clientSocket, WiiMotes ) )
            
if __name__ == '__main__':
    main( '127.0.0.1', 7007, 100 )